@@ -1,15 +1,15 @@
 import asyncio
 import binascii
+from datetime import datetime, timezone
+import enum
 import functools
 import logging
-import enum
-import datetime
 from typing import Any, Dict
 
-import serial
+from zigpy.datastructures import PriorityLock
 import zigpy.exceptions
+import zigpy.types
 
-import zigpy_zigate.config
 import zigpy_zigate.uart
 
 from . import types as t
@@ -39,6 +39,7 @@ class CommandId(enum.IntEnum):
     MANAGEMENT_NETWORK_UPDATE_REQUEST = 0x004A
     SEND_RAW_APS_DATA_PACKET = 0x0530
     AHI_SET_TX_POWER = 0x0806
+    GET_NETWORK_KEY = 0x0054
 
 
 class ResponseId(enum.IntEnum):
@@ -65,8 +66,13 @@ class ResponseId(enum.IntEnum):
     APS_DATA_CONFIRM_FAILED = 0x8702
     AHI_SET_TX_POWER_RSP = 0x8806
     EXTENDED_ERROR = 0x9999
+    GET_NETWORK_KEY_LIST = 0x8054
 
 
+class SendSecurity(t.uint8_t, enum.Enum):
+    NETWORK = 0x00
+    APPLINK = 0x01
+    TEMP_APPLINK = 0x02
 
 
 class NonFactoryNewRestartStatus(t.uint8_t, enum.Enum):
@@ -74,6 +80,7 @@ class NonFactoryNewRestartStatus(t.uint8_t, enum.Enum):
     Running = 1
     Start = 2
 
+
 class FactoryNewRestartStatus(t.uint8_t, enum.Enum):
     Startup = 0
     Start = 2
@@ -137,6 +144,7 @@ RESPONSES = {
     ),
     ResponseId.AHI_SET_TX_POWER_RSP: (t.uint8_t,),
     ResponseId.EXTENDED_ERROR: (t.Status,),
+    ResponseId.GET_NETWORK_KEY_LIST: (zigpy.types.KeyData,),
 }
 
 COMMANDS = {
@@ -207,6 +215,10 @@ class CommandError(zigpy.exceptions.APIException):
     pass
 
 
+class CommandNotSupportedError(CommandError):
+    pass
+
+
 class ZiGate:
     def __init__(self, device_config: Dict[str, Any]):
         self._app = None
@@ -214,8 +226,7 @@ class ZiGate:
         self._uart = None
         self._awaiting = {}
         self._status_awaiting = {}
-        self._lock = asyncio.Lock()
-        self._conn_lost_task = None
+        self._lock = PriorityLock()
 
         self.network_state = None
 
@@ -229,70 +240,29 @@ class ZiGate:
     async def connect(self):
         assert self._uart is None
         self._uart = await zigpy_zigate.uart.connect(self._config, self)
-    
+
     def connection_lost(self, exc: Exception) -> None:
         """Lost serial connection."""
-        LOGGER.warning(
-            "Serial '%s' connection lost unexpectedly: %s",
-            self._config[zigpy_zigate.config.CONF_DEVICE_PATH],
-            exc,
-        )
-        self._uart = None
-        if self._conn_lost_task and not self._conn_lost_task.done():
-            self._conn_lost_task.cancel()
-        self._conn_lost_task = asyncio.ensure_future(self._connection_lost())
-
-    async def _connection_lost(self) -> None:
-        """Reconnect serial port."""
-        try:
-            await self._reconnect_till_done()
-        except asyncio.CancelledError:
-            LOGGER.debug("Cancelling reconnection attempt")
-
-    async def _reconnect_till_done(self) -> None:
-        attempt = 1
-        while True:
-            try:
-                await asyncio.wait_for(self.reconnect(), timeout=10)
-                break
-            except (asyncio.TimeoutError, OSError) as exc:
-                wait = 2 ** min(attempt, 5)
-                attempt += 1
-                LOGGER.debug(
-                    "Couldn't re-open '%s' serial port, retrying in %ss: %s",
-                    self._config[zigpy_zigate.config.CONF_DEVICE_PATH],
-                    wait,
-                    str(exc),
-                )
-                await asyncio.sleep(wait)
-
-        LOGGER.debug(
-            "Reconnected '%s' serial port after %s attempts",
-            self._config[zigpy_zigate.config.CONF_DEVICE_PATH],
-            attempt,
-        )
+        if self._app is not None:
+            self._app.connection_lost(exc)
 
     def close(self):
         if self._uart:
             self._uart.close()
             self._uart = None
-        
-    def reconnect(self):
-        """Reconnect using saved parameters."""
-        LOGGER.debug("Reconnecting '%s' serial port", self._config[zigpy_zigate.config.CONF_DEVICE_PATH])
-        return self.connect()
 
     def set_application(self, app):
         self._app = app
 
     def data_received(self, cmd, data, lqi):
-        LOGGER.debug("data received %s %s LQI:%s", hex(cmd),
-                     binascii.hexlify(data), lqi)
         if cmd not in RESPONSES:
-            LOGGER.warning('Received unhandled response 0x%04x', cmd)
+            LOGGER.warning(
+                "Received unhandled response 0x%04x: %r", cmd, binascii.hexlify(data)
+            )
             return
         cmd = ResponseId(cmd)
         data, rest = t.deserialize(data, RESPONSES[cmd])
+        LOGGER.debug("Response received: %s %s %s (LQI:%s)", cmd, data, rest, lqi)
         if cmd == ResponseId.STATUS:
             if data[2] in self._status_awaiting:
                 fut = self._status_awaiting.pop(data[2])
@@ -303,7 +273,7 @@ class ZiGate:
         self.handle_callback(cmd, data, lqi)
 
     async def wait_for_status(self, cmd):
-        LOGGER.debug('Wait for status to command %s', cmd)
+        LOGGER.debug("Wait for status to command %s", cmd)
 
         if cmd in self._status_awaiting:
             self._status_awaiting[cmd].cancel()
@@ -319,7 +289,7 @@ class ZiGate:
                 del self._status_awaiting[cmd]
 
     async def wait_for_response(self, wait_response):
-        LOGGER.debug('Wait for response %s', wait_response)
+        LOGGER.debug("Wait for response %s", wait_response)
 
         if wait_response in self._awaiting:
             self._awaiting[wait_response].cancel()
@@ -334,8 +304,23 @@ class ZiGate:
                 self._awaiting[wait_response].cancel()
                 del self._awaiting[wait_response]
 
-    async def command(self, cmd, data=b'', wait_response=None, wait_status=True, timeout=COMMAND_TIMEOUT):
-        async with self._lock:
+    def _get_command_priority(self, cmd):
+        return {
+            # Watchdog command is prioritized
+            CommandId.SET_TIMESERVER: 9999,
+            # APS command is deprioritized
+            CommandId.SEND_RAW_APS_DATA_PACKET: -1,
+        }.get(cmd, 0)
+
+    async def command(
+        self,
+        cmd,
+        data=b"",
+        wait_response=None,
+        wait_status=True,
+        timeout=COMMAND_TIMEOUT,
+    ):
+        async with self._lock(priority=self._get_command_priority(cmd)):
             tries = 3
 
             tasks = []
@@ -355,7 +340,9 @@ class ZiGate:
                 tasks.append(status_task)
 
             if wait_response is not None:
-                response_task = asyncio.create_task(self.wait_for_response(wait_response))
+                response_task = asyncio.create_task(
+                    self.wait_for_response(wait_response)
+                )
                 tasks.append(response_task)
 
             try:
@@ -386,23 +373,27 @@ class ZiGate:
                     elif not wait_response and not wait_status:
                         return
             finally:
-                for t in tasks:
-                    if not t.done():
-                        t.cancel()
+                for task in tasks:
+                    if not task.done():
+                        task.cancel()
 
                 await asyncio.gather(*tasks, return_exceptions=True)
 
     async def version(self):
-        return await self.command(CommandId.GET_VERSION, wait_response=ResponseId.VERSION_LIST)
+        return await self.command(
+            CommandId.GET_VERSION, wait_response=ResponseId.VERSION_LIST
+        )
 
     async def version_str(self):
         version, lqi = await self.version()
-        version = '{:x}'.format(version[1])
-        version = '{}.{}'.format(version[0], version[1:])
+        version = "{:x}".format(version[1])
+        version = "{}.{}".format(version[0], version[1:])
         return version
 
     async def get_network_state(self):
-        return await self.command(CommandId.NETWORK_STATE_REQ, wait_response=ResponseId.NETWORK_STATE_RSP)
+        return await self.command(
+            CommandId.NETWORK_STATE_REQ, wait_response=ResponseId.NETWORK_STATE_RSP
+        )
 
     async def set_raw_mode(self, enable=True):
         data = t.serialize([enable], COMMANDS[CommandId.SET_RAWMODE])
@@ -413,21 +404,30 @@ class ZiGate:
         await self.command(CommandId.RESET, wait_response=wait_response)
 
     async def erase_persistent_data(self):
-        await self.command(CommandId.ERASE_PERSISTENT_DATA, wait_status=False, wait_response=ResponseId.PDM_LOADED, timeout=10)
+        await self.command(
+            CommandId.ERASE_PERSISTENT_DATA,
+            wait_status=False,
+            wait_response=ResponseId.PDM_LOADED,
+            timeout=10,
+        )
         await asyncio.sleep(1)
-        await self.command(CommandId.RESET, wait_response=ResponseId.NODE_FACTORY_NEW_RESTART)
+        await self.command(
+            CommandId.RESET, wait_response=ResponseId.NODE_FACTORY_NEW_RESTART
+        )
 
-    async def set_time(self, dt=None):
-        """ set internal time
-        if timestamp is None, now is used
-        """
-        dt = dt or datetime.datetime.now()
-        timestamp = int((dt - datetime.datetime(2000, 1, 1)).total_seconds())
-        data = t.serialize([timestamp], COMMANDS[CommandId.SET_TIMESERVER])
+    async def set_time(self):
+        """set internal time"""
+        timestamp = (
+            datetime.now(timezone.utc) - datetime(2000, 1, 1, tzinfo=timezone.utc)
+        ).total_seconds()
+
+        data = t.serialize([int(timestamp)], COMMANDS[CommandId.SET_TIMESERVER])
         await self.command(CommandId.SET_TIMESERVER, data)
 
     async def get_time_server(self):
-        timestamp, lqi = await self.command(CommandId.GET_TIMESERVER, wait_response=ResponseId.GET_TIMESERVER_LIST)
+        timestamp, lqi = await self.command(
+            CommandId.GET_TIMESERVER, wait_response=ResponseId.GET_TIMESERVER_LIST
+        )
         dt = datetime.datetime(2000, 1, 1) + datetime.timedelta(seconds=timestamp[0])
         return dt
 
@@ -435,14 +435,19 @@ class ZiGate:
         data = t.serialize([enable], COMMANDS[CommandId.SET_LED])
         await self.command(CommandId.SET_LED, data)
 
-    async def set_certification(self, typ='CE'):
-        cert = {'CE': 1, 'FCC': 2}[typ]
+    async def set_certification(self, typ="CE"):
+        cert = {"CE": 1, "FCC": 2}[typ]
         data = t.serialize([cert], COMMANDS[CommandId.SET_CE_FCC])
         await self.command(CommandId.SET_CE_FCC, data)
 
     async def management_network_request(self):
-        data = t.serialize([0x0000, 0x07fff800, 0xff, 5, 0xff, 0x0000], COMMANDS[CommandId.MANAGEMENT_NETWORK_UPDATE_REQUEST])
-        return await self.command(CommandId.MANAGEMENT_NETWORK_UPDATE_REQUEST)#, wait_response=0x804a, timeout=10)
+        data = t.serialize(
+            [0x0000, 0x07FFF800, 0xFF, 5, 0xFF, 0x0000],
+            COMMANDS[CommandId.MANAGEMENT_NETWORK_UPDATE_REQUEST],
+        )
+        return await self.command(
+            CommandId.MANAGEMENT_NETWORK_UPDATE_REQUEST, data
+        )  # , wait_response=0x804a, timeout=10)
 
     async def set_tx_power(self, power=63):
         if power > 63:
@@ -450,14 +455,18 @@ class ZiGate:
         if power < 0:
             power = 0
         data = t.serialize([power], COMMANDS[CommandId.AHI_SET_TX_POWER])
-        power, lqi = await self.command(CommandId.AHI_SET_TX_POWER, data, wait_response=CommandId.AHI_SET_TX_POWER_RSP)
+        power, lqi = await self.command(
+            CommandId.AHI_SET_TX_POWER,
+            data,
+            wait_response=CommandId.AHI_SET_TX_POWER_RSP,
+        )
         return power[0]
 
     async def set_channel(self, channels=None):
         channels = channels or [11, 14, 15, 19, 20, 24, 25, 26]
         if not isinstance(channels, list):
             channels = [channels]
-        mask = functools.reduce(lambda acc, x: acc ^ 2 ** x, channels, 0)
+        mask = functools.reduce(lambda acc, x: acc ^ 2**x, channels, 0)
         data = t.serialize([mask], COMMANDS[CommandId.SET_CHANNELMASK])
         await self.command(CommandId.SET_CHANNELMASK, data)
 
@@ -466,30 +475,58 @@ class ZiGate:
         await self.command(CommandId.SET_EXT_PANID, data)
 
     async def get_devices_list(self):
-        (entries,), lqi = await self.command(CommandId.GET_DEVICES_LIST, wait_response=ResponseId.GET_DEVICES_LIST_RSP)
+        (entries,), lqi = await self.command(
+            CommandId.GET_DEVICES_LIST, wait_response=ResponseId.GET_DEVICES_LIST_RSP
+        )
 
         return list(entries or [])
 
     async def permit_join(self, duration=60):
-        data = t.serialize([0x0000, duration, 1], COMMANDS[CommandId.PERMIT_JOINING_REQUEST])
+        data = t.serialize(
+            [0x0000, duration, 1], COMMANDS[CommandId.PERMIT_JOINING_REQUEST]
+        )
         return await self.command(CommandId.PERMIT_JOINING_REQUEST, data)
 
     async def start_network(self):
-        return await self.command(CommandId.START_NETWORK, wait_response=ResponseId.NETWORK_JOINED_FORMED)
+        return await self.command(
+            CommandId.START_NETWORK, wait_response=ResponseId.NETWORK_JOINED_FORMED
+        )
 
     async def remove_device(self, zigate_ieee, ieee):
-        data = t.serialize([zigate_ieee, ieee], COMMANDS[CommandId.NETWORK_REMOVE_DEVICE])
+        data = t.serialize(
+            [zigate_ieee, ieee], COMMANDS[CommandId.NETWORK_REMOVE_DEVICE]
+        )
         return await self.command(CommandId.NETWORK_REMOVE_DEVICE, data)
 
-    async def raw_aps_data_request(self, addr, src_ep, dst_ep, profile,
-                                   cluster, payload, addr_mode=2, security=0):
-        '''
+    async def raw_aps_data_request(
+        self,
+        addr,
+        src_ep,
+        dst_ep,
+        profile,
+        cluster,
+        payload,
+        addr_mode=t.AddressMode.NWK,
+        security=SendSecurity.NETWORK,
+        radius=0,
+    ):
+        """
         Send raw APS Data request
-        '''
-        radius = 0
-        data = t.serialize([addr_mode, addr,
-                           src_ep, dst_ep, cluster, profile,
-                           security, radius, payload], COMMANDS[CommandId.SEND_RAW_APS_DATA_PACKET])
+        """
+        data = t.serialize(
+            [
+                addr_mode,
+                addr,
+                src_ep,
+                dst_ep,
+                cluster,
+                profile,
+                security,
+                radius,
+                payload,
+            ],
+            COMMANDS[CommandId.SEND_RAW_APS_DATA_PACKET],
+        )
         return await self.command(CommandId.SEND_RAW_APS_DATA_PACKET, data)
 
     def handle_callback(self, *args):
@@ -500,35 +537,12 @@ class ZiGate:
             except Exception as e:
                 LOGGER.exception("Exception running handler", exc_info=e)
 
-    @classmethod
-    async def probe(cls, device_config: Dict[str, Any]) -> bool:
-        """Probe port for the device presence."""
-        api = cls(zigpy_zigate.config.SCHEMA_DEVICE(device_config))
-        try:
-            await asyncio.wait_for(api._probe(), timeout=PROBE_TIMEOUT)
-            return True
-        except (
-            asyncio.TimeoutError,
-            serial.SerialException,
-            zigpy.exceptions.ZigbeeException,
-        ) as exc:
-            LOGGER.debug(
-                "Unsuccessful radio probe of '%s' port",
-                device_config[zigpy_zigate.config.CONF_DEVICE_PATH],
-                exc_info=exc,
-            )
-        finally:
-            api.close()
+    async def get_network_key(self):
+        rsp, _ = await self.command(
+            CommandId.GET_NETWORK_KEY, wait_response=ResponseId.GET_NETWORK_KEY_LIST
+        )
 
-        return False
+        if rsp[0] == t.Status.UnhandledCommand:
+            raise CommandNotSupportedError()
 
-    async def _probe(self) -> None:
-        """Open port and try sending a command"""
-        try:
-            device = next(serial.tools.list_ports.grep(self._config[zigpy_zigate.config.CONF_DEVICE_PATH]))
-            if device.description == 'ZiGate':
-                return
-        except StopIteration:
-            pass
-        await self.connect()
-        await self.set_raw_mode()
+        return rsp[0]
